(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Edg_Impl
imports
  Edg
  Ground_Context_Impl
  Trs_Impl
  QDP_Framework_Impl
begin

no_notation Term.Hole ("\<box>")
notation GCHole ("\<box>")
notation Ground_Context.equiv_class ("[_]")

section {* cap *}

(* an executable version of cap *)
fun capI :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt" where
  "capI _ (Var _) = \<box>"
| "capI R (Fun f ts) = (
    let h = GCFun f (map (capI R) ts) in
    if \<exists>r\<in>set R. Ground_Context.match h (fst r)
      then \<box>
      else h)"

(* and a version for the reversed TRS *)
fun capR :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt" where
  "capR _ (Var _) = \<box>"
| "capR R (Fun f ts) = (
    let h = GCFun f (map (capR R) ts) in
    if \<exists>r\<in>set R. Ground_Context.match h (snd r)
      then \<box>
      else h)"


(* and a version with rule maps *)
fun capRM2 :: "('f \<times> nat \<Rightarrow> ('f, 'v) rules) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) gctxt" where
  "capRM2 _ (Var _) = \<box>"
| "capRM2 rm (Fun f ts) = (
    let h = GCFun f (map (capRM2 rm) ts); n = length ts in
    if \<exists>r\<in>set (rm (f, n)). Ground_Context.match h (fst r)
      then \<box>
      else h)"

lemma capRM2_sound:
  fixes t :: "('f, 'v) term"
  assumes noLeftVars: "\<And>lr. lr \<in> set R \<Longrightarrow> is_Fun (fst lr)"
    and rm: "\<And>f n. set (rm (f, n)) = {(l, r). (l, r) \<in> set R \<and> root l = Some (f, n)}"
  shows "capRM2 rm t = cap (set R) t"
proof (induct t)
  case (Fun f ts)
  let ?hArgs = "(map (cap (set R)) ts) :: ('f, 'v) gctxt list"
  let ?hIArgs = "(map (capRM2 rm) ts) :: ('f, 'v) gctxt list "
  let ?h = "GCFun f ?hArgs"
  let ?hI = "GCFun f ?hIArgs"
  from Fun have heq: "?hI = ?h" by simp
  hence hArgsEq: "?hIArgs = ?hArgs" by simp
  let ?EQ = "\<lambda>(h1::('f, 'v) gctxt) (h2::('f, 'v) gctxt) n.
    (\<exists>r\<in>set R. Ground_Context.match h1 (fst r)) = (\<exists>r\<in>set (rm (f, n)). Ground_Context.match h2 (fst r))"
  have "\<And>h1 h2. h1 = h2 \<Longrightarrow> ?EQ (GCFun f h1) (GCFun f h2) (length h2)"
  proof -
    fix h1 h2 :: "('f, 'v) gctxt list"
    assume h12: "h1 = h2"
    let ?mat = "\<lambda>r. Ground_Context.match (GCFun f h2) (fst r)"
    have "(\<exists>r\<in>set (rm (f, length h2)). ?mat r) = (\<exists>r\<in>set (rm (f, length h2)). ?mat r)" by simp
    also have "\<dots> = (\<exists>r\<in>set R. ?mat r)"
    proof
      assume "\<exists>r\<in>set (rm (f, length h2)). ?mat r"
      from this obtain r where "r \<in> set (rm (f, length h2)) \<and> ?mat r" by auto
      with rm have "r \<in> set R \<and> ?mat r" by auto
      thus "\<exists>r\<in>set R. ?mat r" by auto
    next
      assume "\<exists>r\<in>set R. ?mat r"
      from this obtain r where A: "r \<in> set R \<and> ?mat r" by auto
      with noLeftVars[of r] obtain g ts' rhs where B: "r = (Fun g ts', rhs)" by (cases r, cases "fst r", auto)
      from A B have "f = g \<and> length ts' = length h2" by (cases "f=g", auto simp: Ground_Context.match_def)
      with A B rm have "(Fun f ts', rhs) \<in> set (rm (f, length h2)) \<and> ?mat (Fun f ts', rhs)" by simp
      thus "\<exists>r\<in>set (rm (f, length h2)). ?mat r" by blast
    qed
    thus "?EQ (GCFun f h1) (GCFun f h2) (length h2)" by (simp add: h12)
  qed
  from this[OF hArgsEq[symmetric]] have a: "?EQ ?h ?hI (length ts)" by auto
  thus ?case by (force simp: Let_def heq)
qed simp

definition
  capRM :: "bool \<Rightarrow> (('f \<times> nat) \<Rightarrow> ('f, 'v) rules) \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt"
where
  "capRM nlv rm = (if nlv then capRM2 rm else (\<lambda>t. \<box>))"

lemma capRM: 
  assumes nlv: "nlv = (\<forall>lr\<in>set R. is_Fun (fst lr))"
    and rm: "\<And>f n. set (rm (f, n)) = {(l, r). (l, r) \<in> set R \<and> root l = Some (f, n)}"
  shows "capRM nlv rm = cap (set R)"
proof
  fix t
  show "capRM nlv rm t = cap (set R) t"
  proof (cases nlv)    
    case True
    hence "capRM nlv rm t = capRM2 rm t" by (auto simp: capRM_def)
    also have "... = cap (set R) t"
      by (rule capRM2_sound, insert True[unfolded nlv] rm, auto)      
    finally show ?thesis .
  next
    case False
    hence "capRM nlv rm t = \<box>" by (auto simp: capRM_def)
    also have "... = cap (set R) t"
      by (rule cap_lhs_var[symmetric], insert False[unfolded nlv], auto)
    finally show ?thesis .
  qed  
qed

definition
  capRM_dpp ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> 'dpp \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt"
where
  "capRM_dpp I dpp = capRM (dpp_ops.rules_no_left_var I dpp) (dpp_ops.rules_map I dpp)"

definition
  reverse_capRM_dpp ::
    "('dpp, 'f, 'v) dpp_ops \<Rightarrow> 'dpp \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, 'v) gctxt"
where
  "reverse_capRM_dpp I dpp =
    capRM (dpp_ops.rules_non_collapsing I dpp) (dpp_ops.reverse_rules_map I dpp)"

context dpp_spec
begin

lemma capRM_dpp[simp]:   
  shows "capRM_dpp I d = cap (set (rules d))"
  unfolding capRM_dpp_def
  by (rule capRM, insert rules_no_left_var_sound rules_map_sound, force+)

lemma reverse_capRM_dpp[simp]:   
  shows "reverse_capRM_dpp I d = cap ((set (rules d))^-1)"
  unfolding reverse_capRM_dpp_def reverse_rules[symmetric]
  by (rule capRM, unfold reverse_rules_def,
      unfold rules_non_collapsing_sound reverse_rules_map_sound, force+)
end

lemma capI_sound[simp]:
  fixes R :: "('f,'v) rule list"   
  shows "capI R = cap (set R)"
proof 
  fix t
  show "capI R t = cap (set R) t"
  proof (induct t)
    case (Var x) thus ?case by simp
  next
    case (Fun f ts)
    let ?h = "GCFun f (map (cap (set R)) ts) :: ('f,'v) gctxt"
    let ?hI = "GCFun f (map (capI R) ts) :: ('f,'v) gctxt"
    from Fun have heq: "?hI = ?h" by simp
    let ?EQ = "\<lambda>(h1 :: ('f,'v) gctxt) (h2 :: ('f,'v) gctxt) . ((\<exists> r \<in> set R. Ground_Context.match h1 (fst r)) = (Bex (set R) (\<lambda> r. Ground_Context.match h2 (fst r)) ))"
    have "!! h1 h2 . h1 = h2 \<Longrightarrow> ?EQ h1 h2 "
    proof -
      fix h1 h2 :: "('f,'v)gctxt" 
      show "h1 = h2 \<Longrightarrow> ?EQ h1 h2" by (induct R, auto)
    qed
    hence "?h = ?hI \<Longrightarrow> ?EQ ?h ?hI" .
    with heq have a: "?EQ ?h ?hI" by simp
    show ?case using if_weak_cong[OF a, where x = "GCHole" and y = "GCFun f (map (cap (set R)) ts)",symmetric] unfolding capI.simps cap.simps Let_def heq by force
  qed
qed

lemma capR_sound[simp]:   fixes R :: "('f,'v) rule list"
  shows "capR R = cap ((set R)^-1)"
proof 
  fix t
  let ?swap = "\<lambda> lr. (snd lr, fst lr)"
  have id: "set (map ?swap R) = (set R)^-1" by (induct R, auto)
  have "capR R t = capI (map ?swap R) t"
  proof (induct t)
    case (Fun f ss)
    from Fun have id: "map (capR R) ss = map (capI (map ?swap R)) ss" by auto
    show ?case by (simp add: Let_def id)
  qed simp
  also have "... = cap ((set R)^-1) t" unfolding capI_sound id ..
  finally show "capR R t = cap ((set R)^-1) t" .
qed


section {* EDG *}

fun check_edg :: "('f, 'v) rules \<Rightarrow> (('f, 'v) rule \<times> ('f, 'v) rule) \<Rightarrow> bool" where
  "check_edg R ((s,t), (u,v)) = (Ground_Context.match (capI R t) u \<and> Ground_Context.match (capR R u) t)"

lemma check_edg_sound[simp]: "check_edg R e \<longleftrightarrow> e \<in> EDG (set R)"
  unfolding EDG_def by (cases e) auto

declare check_edg.simps[simp del]

no_notation Ground_Context.GCHole ("\<box>")
no_notation Ground_Context.equiv_class ("[_]")
notation Term.Hole ("\<box>")

end